from inewave.config import MESES_DF, MAX_PATAMARES, MAX_SERIES_SINTETICAS

from cfinterface.components.block import Block
from cfinterface.components.line import Line
from cfinterface.components.field import Field
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.floatfield import FloatField
from typing import List, IO
import numpy as np  # type: ignore
import pandas as pd  # type: ignore


class Submercado(Block):
    """
    Bloco com a informaçao do submercado associado aos valores de Custo
    Marginal de Operação.
    """

    BEGIN_PATTERN = r"CUSTO MARGINAL DE DEMANDA \(\$/MWh\)"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha = Line([LiteralField(12, 70)])

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, Submercado):
            return False
        bloco: Submercado = o
        if not all(
            [
                isinstance(self.data, str),
                isinstance(o.data, str),
            ]
        ):
            return False
        else:
            return self.data == bloco.data

    # Override
    def read(self, file: IO):
        self.data = self.__linha.read(file.readline())[0]


class CmargsAnos(Block):
    """
    Bloco com a informaçao do submercado associado aos valores de Custo
    Marginal de Operação.
    """

    BEGIN_PATTERN = "     ANO: "
    END_PATTERN = "  MEDIA"

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha_ano = Line([IntegerField(4, 10)])
        campos_serie_patamar: List[Field] = [
            IntegerField(4, 2),
            IntegerField(2, 9),
        ]
        campos_custos: List[Field] = [
            FloatField(8, 15 + 9 * i, 2) for i in range(len(MESES_DF) + 1)
        ]
        self.__linha = Line(campos_serie_patamar + campos_custos)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, CmargsAnos):
            return False
        bloco: CmargsAnos = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            cols = ["Série", "Patamar"] + MESES_DF + ["Média"]
            df = pd.DataFrame(tabela, columns=cols)
            df["Ano"] = self.__ano
            df = df[["Ano"] + cols]
            df = df.astype(
                {"Série": "int64", "Patamar": "int64", "Ano": "int64"}
            )
            return df

        self.__ano = self.__linha_ano.read(file.readline())[0]
        file.readline()

        # Variáveis auxiliares
        self.__serie_atual = 0
        tabela = np.zeros(
            (MAX_PATAMARES * MAX_SERIES_SINTETICAS, len(MESES_DF) + 3)
        )
        i = 0
        while True:
            linha = file.readline()
            if self.ends(linha):
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            dados = self.__linha.read(linha)
            if dados[0] is not None:
                self.__serie_atual = dados[0]
            tabela[i, 0] = self.__serie_atual
            tabela[i, 1:] = dados[1:]
            i += 1



from inewave.nwlistop.gttotsin import GttotSIN

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.gttotsin import MockGttotsin


def test_atributos_encontrados_gttotsin():
    m: MagicMock = mock_open(read_data="".join(MockGttotsin))
    with patch("builtins.open", m):
        n = GttotSIN.le_arquivo("")
        assert n.geracao is not None
        assert n.geracao.iloc[0, 0] == 2022
        assert n.geracao.iloc[-1, -1] == 5106.0


def test_atributos_nao_encontrados_gttotsin():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        n = GttotSIN.le_arquivo("")
        assert n.geracao is None


def test_eq_gttotsin():
    m: MagicMock = mock_open(read_data="".join(MockGttotsin))
    with patch("builtins.open", m):
        n1 = GttotSIN.le_arquivo("")
        n2 = GttotSIN.le_arquivo("")
        assert n1 == n2


# Não deve ter teste de diferença, visto que o atributo é
# implementado como Lazy Property.



from inewave.nwlistop.modelos.geolsin import GEAnos

from cfinterface.files.blockfile import BlockFile
import pandas as pd  # type: ignore
from typing import TypeVar, Optional


class GeolSIN(BlockFile):
    """
    Armazena os dados das saídas referentes à geração eólica total
    para o SIN.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `geolsin.out`.
    """

    T = TypeVar("T")

    BLOCKS = [
        GEAnos,
    ]

    def __init__(self, data=...) -> None:
        super().__init__(data)
        self.__gh = None

    @classmethod
    def le_arquivo(
        cls, diretorio: str, nome_arquivo="geolsin.out"
    ) -> "GeolSIN":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="geolsin.out"):
        self.write(diretorio, nome_arquivo)

    def __monta_tabela(self) -> pd.DataFrame:
        df = None
        for b in self.data.of_type(GEAnos):
            dados = b.data
            if dados is None:
                continue
            elif df is None:
                df = b.data
            else:
                df = pd.concat([df, b.data], ignore_index=True)
        return df

    @property
    def geracao(self) -> Optional[pd.DataFrame]:
        """
        Tabela com a geracao eólica por série e
        por mês/ano de estudo.

        :return: A tabela da geração eólica.
        :rtype: Optional[pd.DataFrame]
        """
        if self.__gh is None:
            self.__gh = self.__monta_tabela()
        return self.__gh
